/*
  Serial-t45 (UART and Temp sensor only)
  Onboard temperature with UART(TTL) monitoring at 9600 Baud

  // I included below because I had a problem with the clock and had to reset fuses
  avrdude -p t45 -c usbtiny -P usb -U flash:w:hello.bus.45.c.hex:i
  avrdude -p t45 -c usbtiny -U lfuse:w:0x62:m
  avrdude -p t45 -c usbtiny -U hfuse:w:0xDF:m
  avrdude -p t45 -c usbtiny -U efuse:w:0xFF:m


  
*/

//
//float tlc;
//#define DEBUG     // Comment this out to disable debug scaffold
//#ifdef DEBUG        //// Debug for checking code execution, time since last call in float milliseconds
//  #define db(say) Serial.print(micros()/1000.000-tlc);Serial.print("\t");Serial.println(say);tlc=micros()/1000.000; 
//  #define dbi(insert) insert;db("code insertion") //// Direct insert something
//  #define dbo(instead) instead;db("code override")while(0)   //// Skip a block of code
//#else                            
//  #define db(say)       // Compiles to blank line
//  #define dbi(insert)     
//  #define dbo(instead)    
//#endif

//  db("main()")   // To use debug scaffold e.g.:

#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <string.h>

#define bit_delay() _delay_us(70) // 9600 Baud - my clock is wrong. Neil's delay was 102
#define half_bit_delay() _delay_us(35)  // Sounds obvious but make sure to update this too
#define unit_id 0

void init_prescaler() {
  // Set prescaler division
  CLKPR = (1 << CLKPCE);
  CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (1 << CLKPS1) | (1 << CLKPS0);
}

void init_outputs() {
  // Set pin to high and then output
  PORTB |= (1 << PB4) | (1 << PB0); //PB4 is UART out, PB0 is LED
  DDRB |= (1 << PB4) | (1 << PB0);
}

void triple_flash() {
  int i;
  for (i = 0; i < 3; i++) {
    PORTB &= ~(1 << PB0);
    _delay_ms(100);
    PORTB |= (1 << PB0);
    _delay_ms(100);
  }
}

void serial_rx(volatile unsigned char *pins, unsigned char pin, char *rxbyte) {
  // Read character into rxbyte on PINB
  //        Inverts bits (Assumes line driver)
  *rxbyte = 0;
  // Wait for start bit
  while (*pins & pin) {};
  half_bit_delay(); // Delay half a bit
  int i;
  for (i = 0; i < 8; i++) {
    if (*pins & pin) {
      *rxbyte |= (1 << i);
    } else {
      *rxbyte |= (0 << i);
    }
    half_bit_delay();   // Not sure why, but after checking with a scope the full bit delay is too long
  }
  // Wait for stop bit
  bit_delay(); half_bit_delay();
}

void serial_tx_string(volatile unsigned char *port, unsigned char pin, PGM_P str) {
  // Prepare a string of characters to send through UART TX
  static char chr;
  int i = 0;
  do {
    chr = pgm_read_byte(&(str[i]));
    serial_tx(&PORTB, (1 << PB4), chr);
    ++i;
  } while (chr != 0); 
}

void serial_tx(volatile unsigned char *port, unsigned char pin, char txchar) {
  // Send a character through UART TX 
  //          Inverts bits (Assuming a line driver)
  // Start bit
  *port &= ~pin;
  bit_delay();
  // Write data bits
  int i;
  for (i = 0; i < 8; i++) {
    if (txchar & (1 << i)) {
      *port |= pin;
    } else {
      *port &= ~pin;
    }
    bit_delay();
  }
  // Stop bit
  *port |= pin;
  bit_delay(); 
  // Char delay
  bit_delay();
}

uint16_t adc_lo, adc_hi;

void init_adc() {
  // Initialise ADC
  //ADMUX = (0 << REFS2) | (0 << REFS1) | (0 << REFS0) // Set reference voltage as Vcc
  ADMUX = (0 << REFS2) | (1 << REFS1) | (0 << REFS0) // Set ref. voltage as 1.1V for using ADC4 temp. sensor onboard
          | (0 << ADLAR)                              // No left shift for 10-bit
          //| (0 << MUX3) | (0 << MUX2) | (0 << MUX1) | (1 << MUX0); // PB2/ADC1
          //| (0 << MUX3) | (0 << MUX2) | (1 << MUX1) | (1 << MUX0); // PB3/ADC3
          //| (0 << MUX3) | (0 << MUX2) | (1 << MUX1) | (0 << MUX0); // PB4/ADC2
          //| (0 << MUX3) | (0 << MUX2) | (0 << MUX1) | (0 << MUX0); // PB5/ADC2
          | (1 << MUX3) | (1 << MUX2) | (1 << MUX1) | (1 << MUX0); // ADC4 (Onboard temperature sensor)
                                                                  //Note: 20x gain diff. input and external ref. voltage is available
  ADCSRA = (1 << ADEN)                                // ADC Enable
          | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // Prescaler to 128
}

void read_adc() {
  ADCSRA |= (1 << ADSC);      // Set ADSC bit to 1 to start conversion
  while (ADCSRA & (1 << ADSC)) {}; // Wait until the conversion is complete
  adc_lo = ADCL;  // Read ADCL first
  adc_hi = ADCH;  // Read ADCH 
}

int main(void) {
  // Set up a char for RX pin
  static char chr;
  // Init all the things
  init_prescaler();
  init_outputs();
  init_adc();
  triple_flash();

  // Send data frame
  serial_tx(&PORTB, (1 << PB4), B01010101); // Should come up as U

  while(1) {
    // Read the contents of RX
    serial_rx(&PINB, (1 << PB3), &chr);
    triple_flash();
    if (chr == unit_id) {
      triple_flash(); // Do this again so you can see if interpretation is good
      read_adc();

      // Send a string
      DDRB |= (1 << PB4);   // Prepare for output
      static const char message[] PROGMEM = "ADC4 reads:\t";
      serial_tx_string(&PORTB, (1 << PB4), (PGM_P) message);
      serial_tx(&PORTB, (1 << PB4), &adc_lo);
      serial_tx(&PORTB, (1 << PB4), &adc_hi);

      DDRB &= ~(1 << PB4);
    }
    _delay_ms(10);
  }
  return 1;
}

